<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
  <dict>
    <key>Label</key>
    <string>{{ launchd_service_name }}</string>
    <key>ProgramArguments</key>
    <array>
      <string>/usr/local/sbin/ansible_test_service</string>
      <string>21213</string>
    </array>
  </dict>
</plist>
