# {{ ansible_managed }}
server = true
pid_file = "{{ remote_dir }}/consul.pid"
ports {
  http = 8500
  {% if  pyopenssl_version.stdout is version('0.15', '>=') %}
  https = 8501
  {% endif %}
}
{% if  pyopenssl_version.stdout is version('0.15', '>=') %}
key_file = "{{ remote_dir }}/privatekey.pem"
cert_file = "{{ remote_dir }}/cert.pem"
{% endif %}
